/* February 2020, © Ruud Harmsen, but do with it whatever you see fit.

   SlashExpander takes a multibyte text string, and identifies clusters
   of variants within it, separated by "/" or a specified other string.
   It then recursively alternates the variants, calling a specified
   function for each.

   Example: given the string:
   Aáá Bbb/Ccc/D Eê F/Gg
   the expansion will result in:
   Aáá Bbb Eê F
   Aáá Bbb Eê Gg
   Aáá Ccc Eê F
   Aáá Ccc Eê Gg
   Aáá D Eê F
   Aáá D Eê Gg

   Typically, a program using this will call the three functions
   se_AnalyseString, se_MakeResults and se_FreeAll in succession.
   For further details see above each of the function prototypes
   below.
*/

#ifndef _SLASHEXPANDER_H
#define _SLASHEXPANDER_H

/* In se_SubString, `sub' is a pointer to the start of a substring
   in string `line', and `len' gives its lengths in bytes. Note that
   this can be more than its length in characters, in case the string
   contains multibyte characters, encoded in UTF8 for example, depending
   on the current locale. */
struct se_SubString
{
   char *sub;
   int len;
};
typedef struct se_SubString se_SubstrSpec;

/* The function se_AnalyseString examines the multibyte character
   string `line', and identifies any clusters of one or more
   variants it may contain. Variants are separated by the string
   `separator', which by default (if specified as NULL or "") is "/".

   The function returns a pointer to an array of pointers to
   se_SubString structs, which specify the position and length
   of variants in the clusters found in string `line'.

   For example, if `line' points to:
   Aáá Bbb/Ccc/D Eê F/Gg
   the first pointer in the array points to a one member array
   containing a single struct specifying the position and length
   of string "Aáá". If encoded in UTF8, `len' will be 5, although
   there are only 3 multibyte characters, the last two of which
   occupy two bytes each.

   The second pointer points to a three member array, one specifying
   position and length of "Bbb", the next one being for "Ccc", and
   finally one for "D", then a closing NULL pointer. And so on for
   any further clusters (groups of alphabetic string separated by
   "/") that were found.

   Note that `len' can be zero, in case the line immediately starts
   with variants. So if `line' contains
   Bbb/Ccc/D Eê F/Gg
   the first array member may point to an se_SubstrSpec in which
   `sub' points to the "B" and `len' is 0, whereas the second array
   member is filled as in the previous example. Such an empty cluster
   may also appear at the end.

   The last pointer in the array that the returned pointer points to,
   is always NULL, as is the last pointer in any of the arrays that
   the non-NULL pointers in the array point to. This may be used to
   reliably stop loops.

   The function may return NULL, signalling an allocation error.
 */
se_SubstrSpec **se_AnalyseString (char *line, char *separator);

/* Argument `Places' to function se_MakeResults should be the pointer
   obtained from a previous call of function se_AnalyseString,
   described above.

   Function se_MakeResults recursively expands the clusters of variants
   found in `line'. It calls the specified function *ResFunc for each
   result. The argument Result of that function points to an array of
   se_SubstrSpec, specifying position and byte length of each substring
   of the current expansion.

   A sample function for *ResFunc (see also the test driver in file
   SlashExpander.c) is as follows:

// Sample function start
int ShowResults (se_SubstrSpec *Result)
{
   int cluster;

   for (cluster = 0; Result[cluster].sub; cluster++)
   {
      fprintf(stdout, "%.*s", Result[cluster].len, Result[cluster].sub);
   }

   return 0;
}
// Sample function end

   This prints the assembled, expanded strings to standard output. But of
   course the function can do anything else, based on the specification 
   obtained in argument Result.

   The function se_MakeResults returns a pointer to the internally allocated
   buffer Result (NULL is the allocation fails).
 */
se_SubstrSpec *se_MakeResults (se_SubstrSpec **Places, int (*ResFunc) (se_SubstrSpec *Result));

/* After each call of functions se_AnalyseString and se_MakeResults, se_FreeAll
   should be called with the pointers obtained from those calls. That will free
   the storage that was internally allocated and sized automatically.
 */
void se_FreeAll (se_SubstrSpec **Places, se_SubstrSpec *Result);

#endif
